package net.sf.itcb.addons.monitoring.track.appender;

import java.io.PrintWriter;
import java.io.StringWriter;
import java.io.Writer;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.List;

import javax.wsdl.Types;

import net.sf.itcb.addons.monitoring.track.bean.ItcbTrackBean;
import net.sf.itcb.common.business.core.ItcbProjectInformation;
import net.sf.itcb.common.business.exceptions.BusinessItcbException;

import org.springframework.beans.factory.annotation.Required;
import org.springframework.core.task.TaskExecutor;
import org.springframework.jdbc.core.support.JdbcDaoSupport;

/**
 * Simple DB appender that track in a spring configured database<br/>
 * Tracking in database is done in an asynchronous thread, in order to minimize process impact
 * @author Pierre Le Roux
 */
public class ItcbDbAppender extends JdbcDaoSupport implements ItcbAppender {

	private static final String DONE = "DONE";
	private static final String TECHNICAL_ERROR = "TECHNICAL_ERROR";
	private static final String BUSINESS_ERROR = "BUSINESS_ERROR";

	private TaskExecutor taskExecutor;
	
	private List<String> layersTrackedParamsOnTechnicalError;
	private List<String> layersTrackedParamsOnBusinessError;
	private List<String> layersTrackedParamsOnDoneRequest;
	
	private String module;
	private String moduleVersion;
	
	private Long idModule;
	
	private static final String SQL_INSERT_MODULE =
		"INSERT INTO module (date, module, module_version, server) " +
		"VALUES (now(),?, ?, ?)";
	
	private static final String SQL_INSERT_TRACK =
		"INSERT INTO tracker (date, transaction_id, parent_phase_id, phase_id, user, module_id, layer_type, class, method, time_in_millis, status, exception_type, exception_message, request_message) " +
		"VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)";

	public void setItcbProjectInformation(
			ItcbProjectInformation itcbProjectInformation) {
		this.module = itcbProjectInformation.getGroupId() + "." + itcbProjectInformation.getArtifactId();
		this.moduleVersion = itcbProjectInformation.getVersion();
	}
	
	protected boolean isTrackParamsOnTechnicalError(String layer) {
		return layersTrackedParamsOnTechnicalError.contains(layer);
	}

	public void setLayersTrackedParamsOnTechnicalError(List<String> layersTrackedParamsOnTechnicalError) {
		this.layersTrackedParamsOnTechnicalError = layersTrackedParamsOnTechnicalError;
	}

	protected boolean isTrackParamsOnBusinessError(String layer) {
		return layersTrackedParamsOnBusinessError.contains(layer);
	}

	public void setLayersTrackedParamsOnBusinessError(List<String> layersTrackedParamsOnBusinessError) {
		this.layersTrackedParamsOnBusinessError = layersTrackedParamsOnBusinessError;
	}
	
	protected boolean isTrackParamsOnDoneRequest(String layer) {
		return layersTrackedParamsOnDoneRequest.contains(layer);
	}

	public void setLayersTrackedParamsOnDoneRequest(List<String> layersTrackedParamsOnDoneRequest) {
		this.layersTrackedParamsOnDoneRequest = layersTrackedParamsOnDoneRequest;
	}

	@Required
	public void setTaskExecutor (TaskExecutor taskExecutor) {
		this.taskExecutor=taskExecutor;
	}
	
	
	@Override
	protected void initDao() throws Exception {
		logger.debug("Initialization of Monitoring tracker appender");
		Connection con = null;
		try {
			con = getConnection();
			PreparedStatement stmt = con.prepareStatement(SQL_INSERT_MODULE, Statement.RETURN_GENERATED_KEYS);
			
			stmt.setString(1, module);
			stmt.setString(2, moduleVersion);
			
			String server = java.net.InetAddress.getLocalHost().getHostName();
			stmt.setString(3, server); 
				
			stmt.executeUpdate();
			ResultSet rs = stmt.getGeneratedKeys();
			if(rs.next()) {
				idModule = rs.getLong(1);
				logger.info("This launch of "+ module + " " + moduleVersion + " has been referenced in monitoring database as key " + idModule  );
			}
		} finally {
			releaseConnection(con);
		}
		if(layersTrackedParamsOnTechnicalError == null) {
			layersTrackedParamsOnTechnicalError = new ArrayList<String>();
		}
		if(layersTrackedParamsOnBusinessError == null) {
			layersTrackedParamsOnBusinessError = new ArrayList<String>();
		}
		if(layersTrackedParamsOnDoneRequest == null) {
			layersTrackedParamsOnDoneRequest = new ArrayList<String>();
		}
	}
	
	protected void track(ItcbTrackBean itcbTrackBean) {
		Connection con = null;
		logger.debug("track");
		try {
			con = getConnection();
			PreparedStatement stmt = con.prepareStatement(SQL_INSERT_TRACK);
			
			stmt.setLong(1, itcbTrackBean.getStartTime());
			stmt.setString(2, itcbTrackBean.getTransactionId());
			stmt.setString(3, itcbTrackBean.getParentPhaseId());
			stmt.setString(4, itcbTrackBean.getCompletePhaseId());
			stmt.setString(5, itcbTrackBean.getUser());
			stmt.setLong(6, idModule); 
			stmt.setString(7, itcbTrackBean.getLayerType()); 
			stmt.setString(8, itcbTrackBean.getClassName());
			stmt.setString(9, itcbTrackBean.getMethod());
			stmt.setLong(10, itcbTrackBean.getTimeInMillis());
			Throwable t = itcbTrackBean.getException();
			if( t == null) {
				stmt.setString(11, DONE);
				stmt.setNull(12, Types.STRING_TYPE);
				stmt.setNull(13, Types.STRING_TYPE);
				stmt.setNull(14, Types.STRING_TYPE);
				if(isTrackParamsOnDoneRequest(itcbTrackBean.getLayerType())) {
					logArgs(itcbTrackBean, stmt);
				}
			}
			else  {
				if(t instanceof BusinessItcbException) {
					stmt.setString(11, BUSINESS_ERROR);
					if(isTrackParamsOnBusinessError(itcbTrackBean.getLayerType())) {
						logArgs(itcbTrackBean, stmt);
						stmt.setString(13, getStackTrace(t));
					}
					else {
						stmt.setNull(13, Types.STRING_TYPE);
						stmt.setNull(14, Types.STRING_TYPE);
					}
				}
				else {
					stmt.setString(11, TECHNICAL_ERROR);
					if(isTrackParamsOnTechnicalError(itcbTrackBean.getLayerType())) {
						logArgs(itcbTrackBean, stmt);
						stmt.setString(13, getStackTrace(t));
					}
					else {
						stmt.setNull(13, Types.STRING_TYPE);
						stmt.setNull(14, Types.STRING_TYPE);
						
					}
				}
				stmt.setString(12, t.getClass().getName());
			}
				
			stmt.executeUpdate();
			

		} catch (Exception e) {
			logger.error("An unexpected exception has occurs in traking method : ", e);
		} finally {
			releaseConnection(con);
		}
	}
	
	/**
	 * Log arguments in record
	 * @param itcbTrackBean
	 * @param stmt
	 * @throws SQLException
	 */
	private void logArgs(ItcbTrackBean itcbTrackBean, PreparedStatement stmt)
			throws SQLException {
		if(itcbTrackBean.getParameters() == null) {
			stmt.setNull(14, Types.STRING_TYPE);
		}
		else {
			StringBuilder sb = new StringBuilder();
			for (Object parameter : itcbTrackBean.getParameters()) {
				sb.append(parameter.toString());
				sb.append("\n");
			}
			stmt.setString(14, sb.toString());
		}
	}
	
	/**
	 * Return a stacktrace of throwable
	 * @param aThrowable
	 * @return
	 */
	public static String getStackTrace(Throwable aThrowable) {
	    final Writer result = new StringWriter();
	    final PrintWriter printWriter = new PrintWriter(result);
	    aThrowable.printStackTrace(printWriter);
	    return result.toString();
	}

	@Override
	public void loopTrack(ItcbTrackBean itcbTrackBean) {
		taskExecutor.execute(new AsyncTracker(itcbTrackBean));	
	}
	
	/**
	 * This async mecanism allow to store data in database in an asynchronous thread
	 * 
	 * @author Pierre Le Roux
	 */
	private class AsyncTracker implements Runnable {

		ItcbTrackBean itcbTrackBeanRoot;
		
		public AsyncTracker(ItcbTrackBean itcbTrackBean) {
			this.itcbTrackBeanRoot=itcbTrackBean;
		}

		@Override
		public void run() {
			loopTrack(itcbTrackBeanRoot);
		}
		
		public void loopTrack(ItcbTrackBean itcbTrackBean) {
			track(itcbTrackBean);
			if(itcbTrackBean.getChildren() != null) {
				for (ItcbTrackBean child : itcbTrackBean.getChildren()) {
					loopTrack(child);
				}
			}
		}
		
	}


}
